#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pex',
 'pex.bin',
 'pex.commands',
 'pex.third_party',
 'pex.vendor',
 'pex.vendor._vendored',
 'pex.vendor._vendored.packaging',
 'pex.vendor._vendored.packaging.packaging',
 'pex.vendor._vendored.pip.pip',
 'pex.vendor._vendored.pip.pip._internal',
 'pex.vendor._vendored.pip.pip._internal.cli',
 'pex.vendor._vendored.pip.pip._internal.commands',
 'pex.vendor._vendored.pip.pip._internal.distributions',
 'pex.vendor._vendored.pip.pip._internal.index',
 'pex.vendor._vendored.pip.pip._internal.models',
 'pex.vendor._vendored.pip.pip._internal.network',
 'pex.vendor._vendored.pip.pip._internal.operations',
 'pex.vendor._vendored.pip.pip._internal.operations.build',
 'pex.vendor._vendored.pip.pip._internal.operations.install',
 'pex.vendor._vendored.pip.pip._internal.req',
 'pex.vendor._vendored.pip.pip._internal.utils',
 'pex.vendor._vendored.pip.pip._internal.vcs',
 'pex.vendor._vendored.pip.pip._vendor',
 'pex.vendor._vendored.pip.pip._vendor.cachecontrol',
 'pex.vendor._vendored.pip.pip._vendor.cachecontrol.caches',
 'pex.vendor._vendored.pip.pip._vendor.certifi',
 'pex.vendor._vendored.pip.pip._vendor.chardet',
 'pex.vendor._vendored.pip.pip._vendor.chardet.cli',
 'pex.vendor._vendored.pip.pip._vendor.colorama',
 'pex.vendor._vendored.pip.pip._vendor.distlib',
 'pex.vendor._vendored.pip.pip._vendor.distlib._backport',
 'pex.vendor._vendored.pip.pip._vendor.html5lib',
 'pex.vendor._vendored.pip.pip._vendor.html5lib._trie',
 'pex.vendor._vendored.pip.pip._vendor.html5lib.filters',
 'pex.vendor._vendored.pip.pip._vendor.html5lib.treeadapters',
 'pex.vendor._vendored.pip.pip._vendor.html5lib.treebuilders',
 'pex.vendor._vendored.pip.pip._vendor.html5lib.treewalkers',
 'pex.vendor._vendored.pip.pip._vendor.idna',
 'pex.vendor._vendored.pip.pip._vendor.msgpack',
 'pex.vendor._vendored.pip.pip._vendor.packaging',
 'pex.vendor._vendored.pip.pip._vendor.pep517',
 'pex.vendor._vendored.pip.pip._vendor.pkg_resources',
 'pex.vendor._vendored.pip.pip._vendor.progress',
 'pex.vendor._vendored.pip.pip._vendor.pytoml',
 'pex.vendor._vendored.pip.pip._vendor.requests',
 'pex.vendor._vendored.pip.pip._vendor.urllib3',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.contrib',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.contrib._securetransport',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.packages',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.packages.backports',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.packages.ssl_match_hostname',
 'pex.vendor._vendored.pip.pip._vendor.urllib3.util',
 'pex.vendor._vendored.pip.pip._vendor.webencodings',
 'pex.vendor._vendored.setuptools',
 'pex.vendor._vendored.setuptools.pkg_resources',
 'pex.vendor._vendored.setuptools.pkg_resources._vendor',
 'pex.vendor._vendored.setuptools.pkg_resources._vendor.packaging',
 'pex.vendor._vendored.setuptools.pkg_resources.extern',
 'pex.vendor._vendored.setuptools.setuptools',
 'pex.vendor._vendored.setuptools.setuptools._vendor',
 'pex.vendor._vendored.setuptools.setuptools._vendor.packaging',
 'pex.vendor._vendored.setuptools.setuptools.command',
 'pex.vendor._vendored.setuptools.setuptools.extern',
 'pex.vendor._vendored.wheel',
 'pex.vendor._vendored.wheel.wheel',
 'pex.vendor._vendored.wheel.wheel.cli']

package_data = \
{'': ['*'],
 'pex.vendor._vendored': ['pip/*', 'pip/pip-20.0.dev0.dist-info/*'],
 'pex.vendor._vendored.packaging': ['packaging-19.2.dist-info/*',
                                    'pyparsing-2.4.6.dist-info/*',
                                    'six-1.14.0.dist-info/*'],
 'pex.vendor._vendored.setuptools': ['setuptools-42.0.2.dist-info/*'],
 'pex.vendor._vendored.wheel': ['wheel-0.33.6.dist-info/*'],
 'pex.vendor._vendored.wheel.wheel': ['signatures/*']}

extras_require = \
{"subprocess:python_version<'3'": ['subprocess32>=3.2.7']}

entry_points = \
{'console_scripts': ['pex = pex.bin.pex:main'],
 'distutils.commands': ['bdist_pex = pex.commands.bdist_pex:bdist_pex']}

setup(name='pex',
      version='2.1.14',
      description='The PEX packaging toolchain.',
      author='The PEX developers',
      author_email='pantsbuild@gmail.com',
      url='https://github.com/pantsbuild/pex',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=2.7,<3.9,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
     )
